import click
import json
import os
import time

from pathlib import Path
from watchdog.observers import Observer

from . import (
	__version__ as version,
	__software__ as software
)

from .misc import Module, module_message
from .env import PythonEnvironment
from .watcher import VenvWatcher


context_settings = dict(
	show_default = True,
	ignore_unknown_options = True,
	allow_extra_args = True
)


@click.group('cli')
@click.option('--venv', help='The name of the venv to work with.', default=Path.cwd().name)
@click.version_option(version=version, prog_name=software)
@click.pass_context
def cli(ctx, venv):
	ctx.obj = PythonEnvironment(venv)

	if not ctx.obj.load_config():
		ctx.obj.save_config()


@cli.command('create')
@click.argument('name', required=False, default=None)
@click.option('--nuke', '-n', is_flag=True, help='Remove the existing env directory if it exists')
@click.option('--system', '-s', is_flag=True, help='Allow the venv to use system packages when checking dependencies')
@click.option('--symlink', '-y', is_flag=True, help='Symlink python executable instead of making copies')
@click.pass_context
def cli_create(ctx, name, nuke, system, symlink):
	'Create a new python venv'

	if name:
		ctx.obj.name = name

	ctx.obj.cfg.system_site_packages = system
	ctx.obj.cfg.symlinks = symlink

	click.echo(f'Setting up environment: {ctx.obj.name}')

	if nuke and ctx.obj.path.exists():
		ctx.obj.destroy()

	ctx.obj.create()
	ctx.obj.install_all_modules()
	ctx.obj.save_config()

	click.echo('Successfully created environment :3')


@cli.command('destroy')
@click.pass_context
def cli_destroy(ctx):
	'Remove a python venv'

	if not ctx.obj.path.exists():
		return click.echo('Environment path does not exist')

	if not click.confirm(f'Are you sure you want to remove the venv at {ctx.obj.path}?'):
		return

	ctx.obj.destroy()
	click.echo(f'Destroyed environment: {ctx.obj.name}')


@cli.command('setup')
@click.pass_context
def cli_setup(ctx):
	'Install all modules'

	if not ctx.obj.path.exists():
		ctx.obj.create()

	ctx.obj.install_all_modules()


@cli.command('run', context_settings=context_settings)
@click.argument('command', nargs=-1)
@click.pass_context
def cli_run(ctx, command):
	'Run a command or module in the venv'

	if len(command) == 0:
		return click.echo('Script or module name not provided')

	if Path(command[0]).is_file():
		ctx.obj.run_script(command)

	else:
		ctx.obj.run_module(command)


@cli.command('import')
@click.option('--file', '-f', 'filename', help='The requirements or setuptools cfg file to import', default='requirements.txt')
@click.pass_context
def cli_import(ctx, filename):
	'Import module list from requirements.txt'

	try:
		if filename == 'setup.cfg':
			ctx.obj.import_setuptools()

		else:
			ctx.obj.import_requirements(filename)

	except FileNotFoundError:
		return click.echo(f'Failed to find requirements or setuptools cfg file to import: {filename}')

	except KeyError:
		return click.echo('Missing options.setup_requires in setup.cfg')

	ctx.obj.save_config()
	click.echo(f'Imported requirements file: {filename}')


@cli.command('export')
@click.option('--file', '-f', 'filename', help='The requirements file to import', default='requirements.txt')
@click.pass_context
def cli_export(ctx, filename):
	'Export module list to requirements.txt'

	reqfile = ctx.obj.workdir.joinpath(filename)

	with reqfile.open('w') as fd:
		for module in ctx.obj.modules.values():
			fd.write(module.compile() + '\n')

	click.echo(f'Exported requirements file: {filename}')


@cli.command('edit')
@click.option('--editor', '-e', help='The text editor to use for config editing', default=os.environ.get('EDITOR'))
@click.pass_context
def cli_edit(ctx, editor):
	'Edit pyvenv.cfg file in current directory'

	click.edit(editor=editor, require_save=True, filename=ctx.obj.cfgpath)


@cli.command('open')
@click.pass_context
def cli_open(ctx):
	'Open the venv path in a file browser'

	click.launch('file://' + str(ctx.obj.path))


@cli.command('fix')
@click.pass_context
def cli_fix(ctx):
	'Fetch the version number for any modules without a specifed version'

	for mod in ctx.obj.modules.values():
		version = ctx.obj.module_version(mod.name)

		if not version:
			continue

		mod.version = version

	ctx.obj.save_config()
	cli_list.callback()


@cli.command('list')
@click.pass_context
def cli_list(ctx):
	'List modules to be used in this venv'

	for module in ctx.obj.modules.values():
		installed_version = ctx.obj.module_version(module.name)
		text = module.name

		if module.version:
			text += f'=={module.version}'

		if not installed_version:
			text += f' (not installed)'

		elif module.version != installed_version:
			text += f' (installed: {installed_version})'

		click.echo(text)


@cli.command('add')
@click.argument('modules', nargs=-1)
@click.option('--no-install', '-n', is_flag=True, default=False, help='Do not install the module being added.')
@click.pass_context
def cli_add(ctx, modules, no_install=False):
	'Add a module to the list'

	new_modules = []

	for module in modules:
		mod = ctx.obj.add_module(module)
		mod_version = ctx.obj.module_version(mod)

		if not no_install and mod_version != mod.version:
			ctx.obj.install_module(mod)
			mod.version = ctx.obj.module_version(mod)

		new_modules.append(mod)

	ctx.obj.save_config()

	if len(new_modules) == 1:
		mod = new_modules[0]
		click.echo(module_message('Added module:', mod.name, mod.version))

	else:
		click.echo('Added modules:')

		for mod in new_modules:
			click.echo(module_message('-', mod.name, mod.version))


@cli.command('remove')
@click.argument('modules', nargs=-1)
@click.option('--no-uninstall', '-n', is_flag=True, default=False, help='Do not uninstall the module being removed.')
@click.pass_context
def cli_remove(ctx, modules, no_uninstall=False):
	'Remove a module from the list'

	del_modules = []

	for mod in modules:
		try:
			mod = ctx.obj.remove_module(mod)

		except KeyError:
			continue

		if not no_uninstall and ctx.obj.module_version(mod):
			ctx.obj.uninstall_module(mod)

		del_modules.append(mod)

	ctx.obj.save_config()

	if len(del_modules) == 1:
		mod = del_modules[0]
		click.echo('')
		click.echo(f'Removed module: {mod.name}')

	elif len(del_modules) > 1:
		click.echo('')
		click.echo('Removed modules:')

		for mod in del_modules:
			click.echo(f'- {mod.name}')

	else:
		click.echo('No modules to remove')


@cli.command('install')
@click.argument('modules', nargs=-1)
@click.option('--upgrade', '-u', is_flag=True, default=False, help='If the module is already installed, try to upgrade it.')
@click.pass_context
def cli_install(ctx, modules, upgrade=False):
	'Install a module'

	new_modules = []

	for module in modules:
		mod = Module.new_from_string(module)

		if ctx.obj.module_version(mod):
			continue

		ctx.obj.install_module(mod, upgrade)

		if not mod.version:
			mod.version = ctx.obj.module_version(mod)

		new_modules.append(mod)

	ctx.obj.save_config()

	if len(new_modules) == 1:
		mod = new_modules[0]
		click.echo('')
		click.echo(module_message('Installed module:', mod.name, mod.version))

	elif len(new_modules) > 1:
		click.echo('')
		click.echo('Installed modules:')

		for mod in new_modules:
			click.echo(module_message('-', mod.name, mod.version))

	else:
		click.echo('No modules to install')


@cli.command('uninstall')
@click.argument('modules', nargs=-1)
@click.pass_context
def cli_uninstall(ctx, modules):
	'Uninstall a module'

	del_modules = []

	for module in modules:
		mod = Module.new_from_string(module)

		if not ctx.obj.module_version(mod):
			continue

		ctx.obj.uninstall_module(mod)
		del_modules.append(mod)

	if len(del_modules) == 1:
		mod = del_modules[0]
		click.echo('')
		click.echo(f'Uninstalled module: {mod.name}')

	elif len(del_modules) > 1:
		click.echo('')
		click.echo('Uninstalled modules:')

		for mod in del_modules:
			click.echo(f'- {mod.name}')

	else:
		click.echo('No modules to uninstall')


@cli.group('watcher')
@click.pass_context
def cli_watcher(ctx):
	'Manage the process watcher'


@cli_watcher.command('get')
@click.argument('key', required=False)
@click.pass_context
def cli_watcher_get(ctx, key=None):
	'Check the value of a watcher config option'

	if key:
		if key not in ctx.obj.watcher:
			return click.echo(f'Invalid watcher config option: {key}')

		items = [(key, ctx.obj.watcher[key])]

	else:
		items =  ctx.obj.watcher.items()

	for key, value in items:
		if isinstance(value, list):
			value = ', '.join(value)

		click.echo(f'{key.ljust(14)} = {value}')


@cli_watcher.command('set')
@click.argument('key')
@click.argument('value', nargs=-1, required=False)
@click.pass_context
def cli_watcher_set(ctx, key, value=()):
	'Set a watcher config option'

	if key not in ['command', 'path']:
		return click.echo(f'Invalid watcher config option: {key}')

	value = ' '.join(value)

	try:
		value = ctx.obj.watcher_set_config(key, value)
		ctx.obj.save_config()

	except KeyError:
		return click.echo(f'Invalid watcher config option: {key}')

	click.echo(f'Set watcher config option: {key} = {value}')


@cli_watcher.command('append')
@click.argument('key')
@click.argument('values', nargs=-1, required=True)
@click.pass_context
def cli_watcher_append(ctx, key, values):
	'Append values onto list options'

	if key not in ['ext', 'ignore_dirs', 'ignore_files']:
		return click.echo(f'Invalid watcher config option: {key}')

	if not len(values):
		return click.echo(f'Please provide at least one value')

	for value in values:
		if value not in ctx.obj.watcher[key]:
			ctx.obj.watcher[key].append(value)

	ctx.obj.save_config()

	value_string = ', '.join(ctx.obj.watcher[key])
	click.echo(f'Set watcher config option: {key} = {value_string}')


@cli_watcher.command('append')
@click.argument('key')
@click.argument('values', nargs=-1, required=True)
@click.pass_context
def cli_watcher_append(ctx, key, values):
	'Append values to list options'

	if key not in ['ext', 'ignore_dirs', 'ignore_files']:
		return click.echo(f'Invalid watcher config option: {key}')

	if not len(values):
		return click.echo(f'Please provide at least one value')

	for value in values:
		if value not in ctx.obj.watcher[key]:
			ctx.obj.watcher[key].append(value)

	ctx.obj.save_config()

	value_string = ', '.join(ctx.obj.watcher[key])
	click.echo(f'Set watcher config option: {key} = {value_string}')


@cli_watcher.command('remove')
@click.argument('key')
@click.argument('values', nargs=-1, required=True)
@click.pass_context
def cli_watcher_remove(ctx, key, values):
	'Remove values from list options'

	if key not in ['ext', 'ignore_dirs', 'ignore_files']:
		return click.echo(f'Invalid watcher config option: {key}')

	if not len(values):
		return click.echo(f'Please provide at least one value')

	for value in values:
		try:
			ctx.obj.watcher[key].remove(value)

		except:
			pass

	ctx.obj.save_config()

	value_string = ', '.join(ctx.obj.watcher[key])
	click.echo(f'Set watcher config option: {key} = {value_string}')


@cli_watcher.command('run')
@click.pass_context
def cli_watcher_run(ctx):
	'Watch the directory for changes and restart the specified program'

	handler = VenvWatcher(ctx.obj)

	watcher = Observer()
	watcher.schedule(handler, str(ctx.obj.watcher.path), recursive=True)
	watcher.start()

	handler.run_proc()

	try:
		while True:
			time.sleep(0.1)

	except KeyboardInterrupt:
		pass

	handler.kill_proc()
	watcher.stop()
	watcher.join()


def main():
	cli(prog_name='pyvenv')
