import os

from contextlib import contextmanager

from .misc import DotDict


defaults = dict(
	clear = False,
	prompt = None,
	symlinks = False,
	system_site_packages = False,
	upgrade = True,
	upgrade_deps = True,
	with_pip = True
)


class Config(DotDict):
	def __init__(self, path, data, **kwargs):
		data.update(kwargs)
		super().__init__(defaults)

		self._cfgpath = path.joinpath('pyvenv.cfg')

		for key, value in data.items():
			self[key] = value

		try:
			self.read_cfg()

		except FileNotFoundError:
			pass


	def __setitem__(self, key, value):
		if key not in defaults:
			raise KeyError(f'Invalid config option: {key}')

		if key == 'prompt' and value == '.':
			value = os.path.basename(os.getcwd())

		super().__setitem__(key, value)


	def __delitem__(self, key):
		if key not in defaults:
			raise KeyError(f'Invalid config option: {key}')

		self.__setitem__(key, defaults[key])


	@contextmanager
	def temp_value(self, key, value):
		orig_value = self[key]

		self[key] = value

		try:
			yield self

		finally:
			self[key] = orig_value


	def reset(self):
		self.clear()
		self.update(defaults)


	def read_cfg(self):
		data = {}

		with self._cfgpath.open('r') as fd:
			for line in fd.readlines():
				key, value = line.split('=', 1)
				data[key.strip()] = value.strip()

		return data


	def read(self):
		data = self.read_cfg()

		try: self['system_site_packages'] = data['include-system-site-packages']
		except KeyError: pass

		try: self['symlinks'] = data['symlinks']
		except KeyError: pass

		try: self['with_pip'] = data['with-pip']
		except KeyError: pass


	def write(self):
		data = self.read_cfg()
		data['symlinks'] = str(self.symlinks).lower()
		data['with-pip'] = str(self.with_pip).lower()

		with self._cfgpath.open('w') as fd:
			for key, value in data.items():
				fd.write(f'{key} = {value}\n')
