import json
import shutil
import sys

from configparser import ConfigParser
from pathlib import Path
from venv import EnvBuilder

from .config import Config
from .misc import DotDict, Module, Process


def watcher_defaults():
	return DotDict(
		path = '.',
		command = None,
		env = {},
		ext = ['py', 'pyx', 'pyc'],
		ignore_dirs = ['build', 'config', 'data'],
		ignore_files = ['reload.py', 'test.py', 'pyvenv.py']
	)


class PythonEnvironment(EnvBuilder):
	cfgpath = Path('pyvenv.json').resolve()
	envroot = Path.home().joinpath('.local/share/python-env')
	workdir = Path.cwd()


	def __init__(self, name, config_file=None, envroot=None, **kwargs):
		if config_file:
			self.cfgpath = Path(config_file or envroot)

		if envroot:
			self.envroot = Path(envroot)

		self.name = name
		self.modules = DotDict()
		self.watcher = watcher_defaults()
		self.watcher['command'] = name
		self.cfg = Config(self.path, kwargs)


	@property
	def path(self):
		return self.envroot.joinpath(self.name)


	@property
	def executable(self):
		return self.path.joinpath('bin/python')


	## Use Config class for properties normally set by __init__
	@property
	def clear(self):
		return self.cfg['clear']


	@property
	def prompt(self):
		return self.cfg.prompt


	@property
	def symlinks(self):
		return self.cfg.symlinks


	@property
	def system_site_packages(self):
		return self.cfg.system_site_packages


	@property
	def upgrade(self):
		return self.cfg.upgrade


	@property
	def upgrade_deps(self):
		return self.cfg.upgrade_deps


	@property
	def upgrade_deps(self):
		return self.cfg.upgrade_deps


	@property
	def with_pip(self):
		return self.cfg.with_pip


	## EnvBuilder tries to set this
	@system_site_packages.setter
	def system_site_packages(self, value):
		pass


	def add_module(self, mod):
		if not isinstance(mod, Module):
			mod = Module.new_from_string(mod)

		if not mod.version:
			mod.version = self.module_version(mod.name)

		self.modules[mod.name] = mod
		return mod


	def create(self):
		with self.cfg.temp_value('system_site_packages', False):
			super().create(self.path)


	def create_script(self):
		script = Path('python')

		if script.is_symlink():
			return False

		with Path('python').open('w') as fd:
			fd.write(f'#!/bin/sh\n')
			fd.write(f'{self.executable} $@')

		return True


	def destroy(self):
		if self.path:
			shutil.rmtree(self.path)


	def get_module(self, mod):
		if not isinstance(mod, Module):
			mod = Module.new_from_string(mod)

		return self.modules[mod.name]


	def import_requirements(self, filename='requirements.txt'):
		reqfile = Path(filename)

		if not reqfile.exists():
			raise FileNotFoundError(filename)

		with reqfile.open('r') as fd:
			for line in fd.readlines():
				if line.strip() == '.':
					self.import_setuptools()
					continue

				mod = Module.new_from_string(line)
				ctx.obj.modules[mod.name] = mod


	def import_setuptools(self, filename='setup.cfg'):
		cfgfile = Path(filename)

		if not cfgfile.exists():
			raise FileNotFoundError(filename)

		with cfgfile.open('r') as fd:
			cfg = ConfigParser()
			cfg.read_file(fd)

		requirements = ''

		try:
			requirements += cfg['options']['install_requires']

		except KeyError:
			pass

		try:
			requirements += '\n' + cfg['options']['setup_requires']

		except:
			pass

		if not requirements:
			raise KeyError('Missing options.install_requires or options.setup_requires')

		for line in requirements.splitlines():
			if not line:
				continue

			self.add_module(line)


	def install_all_modules(self):
		for module in self.modules.values():
			installed_version = self.module_version(module.name)

			if not installed_version:
				self.install_module(module)

			elif module.version and module.version != installed_version:
				self.install_module(module, update=True)


	def install_module(self, mod, update=False):
		if isinstance(mod, str):
			mod = Module.new_from_string(mod)

		env = {}

		if self.system_site_packages:
			env['SETUPTOOLS_USE_DISTUTILS'] = 'stdlib'

		command = ['pip', 'install', mod.compile(False)]

		if update:
			command.insert(2, '-U')

		return self.run_module(command, env=env)


	def load_config(self):
		try:
			with self.cfgpath.open('r') as fd:
				config = json.load(fd)

		except FileNotFoundError:
			return False

		self.name = config['name']

		for name, info in config.get('modules', {}).items():
			self.modules[name] = Module(name, **info)

		for key, value in config.get('watcher', {}).items():
			if key in self.watcher:
				self.watcher[key] = value

		return True


	def module_version(self, mod):
		if isinstance(mod, Module):
			name = mod.name

		else:
			name = mod

		exec_string = f'from importlib.metadata import distribution; print(distribution("{name}").version)'

		proc = Process(f'{self.executable} -c \'{exec_string}\'', capture_output=True)
		return proc.stdout.strip()


	def remove_module(self, mod):
		if not isinstance(mod, Module):
			mod = Module.new_from_string(mod)

		if not self.modules.get(mod.name):
			raise KeyError(f'module "{mod.name}" not in module list')

		del self.modules[mod.name]
		return mod


	def run_module(self, module, **kwargs):
		if isinstance(module, (list, tuple)):
			module = ['-m', *module]

		elif isinstance(module, str):
			module = f'-m {module}'

		else:
			raise TypeError(f'command must be a list, tuple, or str, not {type(module).__name__}')

		return self.run_python(module, **kwargs)


	def run_python(self, command, **kwargs):
		if not self.executable.exists():
			raise FileNotFoundError('Cannot find python executable. Was the environment created?')

		if isinstance(command, (list, tuple)):
			command = [self.executable, *command]

		elif isinstance(command, str):
			command = f'{self.executable} {command}'

		else:
			raise TypeError(f'command must be a list, tuple, or str, not {type(command).__name__}')

		return Process(command, **kwargs)


	def run_script(self, script, **kwargs):
		return self.run_python(script, **kwargs)


	def save_config(self):
		config = dict(
			name = self.name,
			modules = dict(sorted(self.modules.items())),
			watcher = self.watcher
		)

		with self.cfgpath.open('w') as fd:
			json.dump(config, fd, indent='\t')

		try:
			self.cfg.write()

		except FileNotFoundError:
			pass


	def uninstall_module(self, mod):
		if not isinstance(mod, Module):
			mod = Module.new_from_string(mod)

		return self.run_module(['pip', 'uninstall', '-y', mod.name])


	def watcher_set_config(self, key, value=None):
		if key not in self.watcher:
			raise KeyError(f'Invalid watcher config option: {key}')

		if not value:
			self.watcher[key] = watcher_defaults()[key]

		else:
			self.watcher[key] = value

		return self.watcher[key]
