import os
import shlex
import signal
import subprocess
import time

import pkg_resources as pkg


def module_message(message, name, version=None):
	return f'{message} {name}{" v" + version if version else ""}'


def run_command(command, *args, **kwargs):
	assert isinstance(command, (str, tuple, list))

	if 'check' not in kwargs:
		kwargs['check'] = True

	if 'capture_output' not in kwargs:
		kwargs['capture_output'] = True

		if 'text' not in kwargs:
			kwargs['text'] = True

	return subprocess.run(shlex.split(command) if isinstance(command, str) else command, *args, **kwargs)


def signal_handler(func, *args, original_args=False, **kwargs):
	if func:
		if original_args:
			handler = lambda signum, frame: func(signum, frame, *args, **kwargs)

		else:
			handler = lambda *_: func(*args, **kwargs)

	else:
		handler = signal.SIG_DFL

	signal.signal(signal.SIGHUP, handler)
	signal.signal(signal.SIGINT, handler)
	signal.signal(signal.SIGQUIT, handler)
	signal.signal(signal.SIGTERM, handler)


class DotDict(dict):
	def __setitem__(self, key, value):
		if type(value) == dict:
			value = DotDict(value)

		super().__setitem__(key, value)


	def __getattr__(self, key):
		try:
			return self[key]

		except KeyError:
			return super().__getattribute__(key)


	def __setattr__(self, key, value):
		if key.startswith('_'):
			return super().__setattr__(key, value)

		self[key] = value


	def __delattr__(self, key, value):
		try:
			del self[key]

		except KeyError:
			super().__delattr__(key)


# todo: support more specifiers than "=="
class Module(DotDict):
	__slots__ = ['version', 'options', 'url']

	def __init__(self, name, version=None, options=[], url=None):
		super().__init__(
			version = version,
			options = options,
			url = url
		)

		self._name = name


	def __str__(self):
		return self.compile()


	def __repr__(self):
		text = f'Module({self.name}'

		if len(self.options):
			opts = ','.join(self.options)
			text += f', options=[{opts}]'

		if self.version:
			text += f', version={self.version}'

		if self.url:
			text += f', url={self.url}'

		text += ')'

		return text


	@classmethod
	def new_from_string(cls, string):
		try:
			module = list(pkg.parse_requirements(string))[0]
		except IndexError:
			return

		try: version = module.specs[0][1]
		except IndexError: version = None

		return cls(
			module.key,
			version,
			list(module.extras),
			module.url
		)


	@property
	def name(self):
		return self._name.lower().replace('_', '-')


	@name.setter
	def name(self, value):
		self._name = value


	def compile(self, with_version=True):
		text = self.name

		if self.options:
			optline = ','.join(self.options)
			text += f'[{optline}]'

		if self.url:
			text += f'@{self.url}'

		elif self.version and with_version:
			text += f'=={self.version}'

		return text


class Process:
	def __init__(self, command, *args, run=True, **kwargs):
		if isinstance(command, str):
			self.command = shlex.split(command)

		elif isinstance(command, (tuple, list)):
			self.command = command

		else:
			raise TypeError(f'Command must be a str, tuple, or list, not {type(command).__name__}')

		self.args = args
		self.kwargs = kwargs
		self._proc = None

		if run:
			self.run()


	def __enter__(self):
		self.run()
		return self


	def __exit__(self, *args):
		pass


	@property
	def exitcode(self):
		return self._proc.returncode


	@property
	def stdout(self):
		try:
			return self._proc.stdout.decode('utf-8')

		except AttributeError:
			return


	@property
	def stderr(self):
		try:
			return self._proc.stdout.decode('utf-8')

		except AttributeError:
			return


	def readlines(self, pipe='stdout', text=True):
		if pipe not in ['stdout', 'stderr']:
			raise ValueError(f'Can only read "stdout" or "stderr", not {pipe}')

		if not self._proc:
			return

		for line in getattr(self, pipe).splitlines():
			yield line


	def run(self):
		if self._proc:
			return

		kwargs = {key: value for key, value in self.kwargs.items() if key != 'env'}
		env = None

		if 'env' in kwargs:
			env = os.environ.copy()
			env.update(self.kwargs['env'])

		self._proc = subprocess.run(self.command, *self.args, env=env, **kwargs)


class ProcessHandler:
	def __init__(self, command, *args, **kwargs):
		self.proc = None
		self.command = shlex.split(command)
		self.args = args
		self.kwargs = kwargs

		if 'shell' not in self.kwargs:
			self.kwargs['shell'] = True

		self._end_proc = False


	def __enter__(self):
		self.start()
		return self


	def __exit__(self, *args, **kwargs):
		self.stop()


	@property
	def exitcode(self):
		try: return self.proc.poll()
		except AttributeError: 0


	@property
	def stdout(self):
		return self.proc.stdout


	@property
	def stderr(self):
		return self.proc.stderr


	def start(self):
		if self.proc:
			return

		signal_handler(self.stop)
		self._end_proc = False
		self.proc = subprocess.Popen(self.command, *self.args, **self.kwargs)


	def run(self):
		if not self.proc:
			self.start()

		while True:
			if self.exitcode != None or self._end_proc:
				break

			time.sleep(0.1)

		return self.exitcode


	def stop(self):
		if not self.proc:
			return

		self._end_proc = True
		self.proc.terminate()

		count = 0

		while self.exitcode == None:
			count += 1
			time.sleep(0.1)

			if count >= 10:
				self.proc.kill()
				break

		signal_handler(None)
		self.proc = None


	def read(self, pipe='stdout', text=True):
		if pipe.lower() not in ['stdout', 'stderr']:
			raise ValueError(f'Can only read "stdout" or "stderr", not {pipe}')

		if not self.proc:
			return

		data = self.proc.stdout.read() if pipe == 'stdout' else self.proc.stderr.read()

		if text:
			try:
				return data.decode('utf-8').strip()

			except:
				pass

		return data
